#lang play

(require "env.rkt")

#|
  <expr>  ::= <num>
            | <sym>
            | <bool>
            | {if <expr> <expr> <expr>}
            | {+ <expr> <expr>}
            | {< <expr> <expr>}
            | {* <expr> <expr>}
            | {= <expr> <expr>}
            | {- <expr> <expr>}
            | {and <expr> <expr>}
            | {or <expr> <expr>}
            | {not <expr> <expr>}
            | {begin <expr> <expr>}
            | {with {<binding>*} <expr>}

<binding> ::= {<sym> <expr>}

; EXTENSIÓN PARA CLASE Y OBJETOS
 <expr>  ::= ... (expresiones del lenguaje entregado) ...
           | {class {<sym>*} <method>*}
           | {new <expr> <expr>*}
           | {get <expr> <sym>}
           | {set <sym> <expr>}
           | {-> <expr> <sym> <expr>*}
           | self

<method> ::= {def <sym> {<sym>*} <expr>}

; EXTENSIÓN PARA LAMBDAS
; SÓLO COMO AZÚCAR SINTÁCTICA. NO DEBEN MODIFICAR AST NI INTERP.
 <expr>  ::= ... (lenguaje base + clases y objetos ) ...
           | {fun {<sym>*} <expr>}
           | {<expr> <expr>*}
|#


(deftype Expr
  (num n)
  (bool b)
  (id s)   
  (binop f l r)
  (unop f s)
  (my-if c tb fb)  
  (begn expr1 expr2)  
  (with defs body))

(deftype Binding
  (binding id expr))

;; values
(deftype Val
  (numV n)
  (boolV b))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; parse :: s-expr -> Expr
(define (parse s-expr)
  (match s-expr
    [(? number?) (num s-expr)]
    [(? symbol?) (id s-expr)]    
    [(? boolean?) (bool s-expr)]
    [(list '* l r) (binop * (parse l) (parse r))]
    [(list '+ l r) (binop + (parse l) (parse r))]
    [(list '- l r) (binop - (parse l) (parse r))]
    [(list '< l r) (binop < (parse l) (parse r))]
    [(list '= l r) (binop = (parse l) (parse r))]    
    [(list 'or l r) (binop (λ (i d) (or i d)) (parse l) (parse r))]
    [(list 'and l r) (binop (λ (i d) (and i d)) (parse l) (parse r))]
    [(list 'not b) (unop not (parse b))]
    [(list 'if c t f) (my-if (parse c)
                             (parse t)
                             (parse f))]
    [(list 'begin e1 e2) (begn (parse e1) (parse e2))]    
    [(list 'with (list e ...)  b)
     (with (map parse-binding e) (parse b))]
    ))


;; parse-binding :: s-expr -> Def
(define (parse-binding s-expr)
  (match s-expr
    [(list id b) (binding id (parse b))]))


;; well-formed :: Expr -> Boolean or error
;; -- COMPLETAR --
;; -- Esto es sólo un esqueleto base --
(define (well-formed expr)
  (match expr
    [(id x) #t]
    [(binop _ l r) (begin (well-formed l) (well-formed r))]
    [(unop _ s) (well-formed s)]
    [(my-if c t f) (for-each well-formed (list c t f))]
    [(begn e1 e2) (begin (well-formed e1) (well-formed e2))]
    [(with defs body)
     (begin (for-each well-formed (map binding-expr defs))
            (well-formed body))]
    [_ #t]))


;; interp :: Expr Env -> Val
(define (interp expr env)
  (match expr
    [(num n) (numV n)]    
    [(bool b) (boolV b)]    
    [(binop f l r) (make-val (f (open-val (interp l env))
                                (open-val (interp r env))))]
    [(unop f s) (make-val (f (open-val (interp s env))))]
    [(my-if c t f)
     (def (boolV cnd) (interp c env))
     (if cnd
         (interp t env)
         (interp f env))]
    [(id x) (env-lookup x env)]        
    [(begn expr1 expr2) (begin 
                          (interp expr1 env)
                          (interp expr2 env))]
    [(with defs body)
     (interp body
             (foldl (λ (x acc-env)
                      (def (binding id expr) x)
                      (extend-env id (interp expr acc-env) acc-env))
                    env defs))     
     ]))

;; open-val :: Val -> Scheme Value
(define (open-val v)
  (match v
    [(numV n) n]
    [(boolV b) b]
    ))

;; make-val :: Scheme Value -> Val
(define (make-val v)
  (match v
    [(? number?) (numV v)]
    [(? boolean?) (boolV v)]
    ))

;; run :: s-expr -> Val
(define (run s-expr)
  (interp (parse s-expr) empty-env))

#|
run-val:: s-expr -> Scheme-Val + Val
Versión alternativa de run, que retorna valores de scheme para primitivas y
valores de MiniScheme para clases y objetos
|#
(define (run-val s-expr)
  (def parsed-expr (parse s-expr))
  (begin
    (well-formed parsed-expr)
    (match (interp parsed-expr empty-env)
      [(numV n) n]
      [(boolV b) b]
      [x x])))